"""Tests for data loading functionality"""
import pytest
from pathlib import Path
from io import StringIO
from pandoc_embedz.data_loader import load_data, guess_format_from_filename

FIXTURES_DIR = Path(__file__).parent / 'fixtures'


class TestGuessFormat:
    """Tests for format auto-detection"""

    def test_csv_extension(self):
        assert guess_format_from_filename('data.csv') == 'csv'

    def test_tsv_extension(self):
        assert guess_format_from_filename('data.tsv') == 'tsv'

    def test_json_extension(self):
        assert guess_format_from_filename('data.json') == 'json'

    def test_yaml_extension(self):
        assert guess_format_from_filename('data.yaml') == 'yaml'
        assert guess_format_from_filename('data.yml') == 'yaml'

    def test_toml_extension(self):
        assert guess_format_from_filename('data.toml') == 'toml'

    def test_sqlite_extensions(self):
        assert guess_format_from_filename('data.db') == 'sqlite'
        assert guess_format_from_filename('data.sqlite') == 'sqlite'
        assert guess_format_from_filename('data.sqlite3') == 'sqlite'

    def test_txt_extension(self):
        assert guess_format_from_filename('data.txt') == 'lines'

    def test_default_to_csv(self):
        assert guess_format_from_filename('data.unknown') == 'csv'


class TestLoadCSV:
    """Tests for CSV data loading"""

    def test_load_csv_with_header(self):
        data = load_data(str(FIXTURES_DIR / 'sample.csv'), format='csv', has_header=True)
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'
        assert data[0]['value'] == 42
        assert data[0]['category'] == 'A'

    def test_load_csv_without_header(self):
        data = load_data(str(FIXTURES_DIR / 'sample.csv'), format='csv', has_header=False)
        assert len(data) == 4  # Including header row as data
        assert isinstance(data[0], list)

    def test_load_csv_inline(self):
        csv_data = StringIO("name,value\nArthur,100\nFord,85")
        data = load_data(csv_data, format='csv', has_header=True)
        assert len(data) == 2
        assert data[0]['name'] == 'Arthur'

    def test_load_csv_with_query(self):
        """Test CSV with SQL query filter"""
        data = load_data(
            str(FIXTURES_DIR / 'sample.csv'),
            format='csv',
            query='SELECT * FROM data WHERE category = "A"'
        )
        assert len(data) == 2
        assert data[0]['name'] == 'Arthur'
        assert data[1]['name'] == 'Zaphod'

    def test_load_csv_with_aggregation_query(self):
        """Test CSV with SQL aggregation"""
        data = load_data(
            str(FIXTURES_DIR / 'sample.csv'),
            format='csv',
            query='SELECT category, COUNT(*) as count, AVG(value) as avg_value FROM data GROUP BY category'
        )
        assert len(data) == 2
        # Category A has 2 items (Arthur:42, Zaphod:99), avg = 70.5
        # Category B has 1 item (Ford:100), avg = 100
        cat_a = [row for row in data if row['category'] == 'A'][0]
        assert cat_a['count'] == 2
        assert cat_a['avg_value'] == 70.5

    def test_load_csv_with_order_by_query(self):
        """Test CSV with SQL ORDER BY"""
        data = load_data(
            str(FIXTURES_DIR / 'sample.csv'),
            format='csv',
            query='SELECT * FROM data ORDER BY value DESC'
        )
        assert len(data) == 3
        assert data[0]['name'] == 'Ford'  # value = 100
        assert data[1]['name'] == 'Zaphod'  # value = 99
        assert data[2]['name'] == 'Arthur'  # value = 42


class TestCSVComments:
    """Tests for CSV comment handling"""

    def test_csv_comment_line_default(self):
        """Default: all # lines are stripped"""
        csv_data = StringIO("# auto-generated\nname,value\n# mid comment\nAlice,100\n")
        data = load_data(csv_data, format='csv')
        assert len(data) == 1
        assert data[0]['name'] == 'Alice'

    def test_csv_comment_head(self):
        """comment=head preserves # in data rows"""
        csv_data = StringIO("# comment\nname,value\n#tag,100\n")
        data = load_data(csv_data, format='csv', comment='head')
        assert len(data) == 1
        assert data[0]['name'] == '#tag'

    def test_csv_comment_none(self):
        """comment=none disables comment handling"""
        csv_data = StringIO("# comment,x\nname,value\nAlice,100\n")
        data = load_data(csv_data, format='csv', has_header=False, comment='none')
        assert len(data) == 3
        assert data[0][0] == '# comment'

    def test_csv_comment_line(self):
        """comment=line strips all # lines"""
        csv_data = StringIO("# header comment\nname,value\n# mid comment\nAlice,100\n")
        data = load_data(csv_data, format='csv', comment='line')
        assert len(data) == 1
        assert data[0]['name'] == 'Alice'

    def test_csv_comment_inline(self):
        """comment=inline strips from # to end of line"""
        csv_data = StringIO("name,value\nAlice,100 # note\n")
        data = load_data(csv_data, format='csv', comment='inline')
        assert len(data) == 1
        assert data[0]['value'] == 100

    def test_csv_comment_false(self):
        """comment=False disables comment handling"""
        csv_data = StringIO("# data,x\nname,value\n")
        data = load_data(csv_data, format='csv', has_header=False, comment=False)
        assert len(data) == 2
        assert data[0][0] == '# data'

    def test_tsv_comment_head(self):
        """Comment handling works with TSV"""
        tsv_data = StringIO("# comment\nname\tvalue\nAlice\t100\n")
        data = load_data(tsv_data, format='tsv')
        assert len(data) == 1
        assert data[0]['name'] == 'Alice'


class TestLoadTSV:
    """Tests for TSV data loading"""

    def test_load_tsv_with_header(self):
        data = load_data(str(FIXTURES_DIR / 'sample.tsv'), format='tsv', has_header=True)
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'
        assert data[1]['name'] == 'Ford'

    def test_load_tsv_inline(self):
        tsv_data = StringIO("name\tvalue\nArthur\t42\nFord\t100")
        data = load_data(tsv_data, format='tsv', has_header=True)
        assert len(data) == 2
        assert data[0]['value'] == 42

    def test_load_tsv_with_spaces_in_values(self):
        """Test that TSV correctly handles spaces within field values"""
        tsv_data = StringIO("name\tvalue\tcomment\nArthur Dent\t42\tHoopy frood\nFord Prefect\t100\tGreat guy")
        data = load_data(tsv_data, format='tsv', has_header=True)

        assert len(data) == 2
        # Spaces within values should be preserved
        assert data[0]['name'] == 'Arthur Dent'
        assert data[0]['comment'] == 'Hoopy frood'
        assert data[1]['name'] == 'Ford Prefect'
        assert data[1]['comment'] == 'Great guy'

    def test_load_tsv_without_header(self):
        """Test TSV without header row"""
        tsv_data = StringIO("Arthur Dent\t42\nFord Prefect\t100")
        data = load_data(tsv_data, format='tsv', has_header=False)

        assert len(data) == 2
        assert data[0][0] == 'Arthur Dent'
        assert data[0][1] == 42


class TestLoadSSV:
    """Tests for SSV (space-separated) data loading"""

    def test_load_ssv_with_header(self):
        ssv_data = StringIO("name value category\nArthur 42 A\nFord 100 B")
        data = load_data(ssv_data, format='ssv', has_header=True)
        assert len(data) == 2
        assert data[0]['name'] == 'Arthur'
        assert data[0]['value'] == 42

    def test_load_ssv_without_header(self):
        ssv_data = StringIO("Arthur 42 A\nFord 100 B")
        data = load_data(ssv_data, format='ssv', has_header=False)
        assert len(data) == 2
        assert isinstance(data[0], list)
        assert data[0][0] == 'Arthur'

    def test_load_spaces_alias_with_header(self):
        """Test that 'spaces' is an alias for 'ssv' with header"""
        spaces_data = StringIO("name value category\nArthur 42 A\nFord 100 B")
        data = load_data(spaces_data, format='spaces', has_header=True)
        assert len(data) == 2
        assert data[0]['name'] == 'Arthur'
        assert data[0]['value'] == 42

    def test_load_spaces_alias_without_header(self):
        """Test that 'spaces' is an alias for 'ssv' without header"""
        spaces_data = StringIO("Arthur 42 A\nFord 100 B")
        data = load_data(spaces_data, format='spaces', has_header=False)
        assert len(data) == 2
        assert isinstance(data[0], list)
        assert data[0][0] == 'Arthur'

    def test_load_ssv_ragged_rows_strip_nan(self):
        """Test that ragged rows (unequal column count) don't produce NaN"""
        ssv_data = StringIO("a b c d e f\ng h i j\nk l")
        data = load_data(ssv_data, format='ssv', has_header=False)
        assert len(data) == 3
        assert data[0] == ['a', 'b', 'c', 'd', 'e', 'f']
        assert data[1] == ['g', 'h', 'i', 'j']
        assert data[2] == ['k', 'l']

    def test_load_csv_ragged_rows_strip_nan(self):
        """Test that ragged CSV rows don't produce NaN"""
        csv_data = StringIO("a,b,c\nd,e\nf")
        data = load_data(csv_data, format='csv', has_header=False)
        assert len(data) == 3
        assert data[0] == ['a', 'b', 'c']
        assert data[1] == ['d', 'e']
        assert data[2] == ['f']

    def test_load_ssv_with_columns_preserves_spaces_in_last_column(self):
        """Test that columns parameter preserves spaces in last column"""
        ssv_data = StringIO("ID Name Description\n1 Alice Software engineer\n2 Bob Project manager with team")
        data = load_data(ssv_data, format='ssv', has_header=True, columns=3)
        assert len(data) == 2
        assert data[0]['ID'] == '1'
        assert data[0]['Name'] == 'Alice'
        assert data[0]['Description'] == 'Software engineer'
        assert data[1]['ID'] == '2'
        assert data[1]['Name'] == 'Bob'
        assert data[1]['Description'] == 'Project manager with team'

    def test_load_ssv_with_columns_without_header(self):
        """Test columns parameter without header row"""
        ssv_data = StringIO("1 Alice Software engineer\n2 Bob Project manager with team")
        data = load_data(ssv_data, format='ssv', has_header=False, columns=3)
        assert len(data) == 2
        assert data[0] == ['1', 'Alice', 'Software engineer']
        assert data[1] == ['2', 'Bob', 'Project manager with team']

    def test_load_ssv_with_columns_multiple_spaces(self):
        """Test columns parameter handles multiple consecutive spaces"""
        ssv_data = StringIO("ID  Name  Description\n1   Alice   Has   multiple   spaces")
        data = load_data(ssv_data, format='ssv', has_header=True, columns=3)
        assert len(data) == 1
        assert data[0]['ID'] == '1'
        assert data[0]['Name'] == 'Alice'
        assert data[0]['Description'] == 'Has   multiple   spaces'

    def test_load_ssv_with_columns_fewer_fields(self):
        """Test columns parameter pads with empty strings when fewer fields"""
        ssv_data = StringIO("ID Name Description\n1 Alice\n2 Bob Complete")
        data = load_data(ssv_data, format='ssv', has_header=True, columns=3)
        assert len(data) == 2
        assert data[0]['ID'] == '1'
        assert data[0]['Name'] == 'Alice'
        assert data[0]['Description'] == ''
        assert data[1]['Description'] == 'Complete'

    def test_load_ssv_with_columns_empty_input(self):
        """Test columns parameter with empty input"""
        ssv_data = StringIO("")
        data = load_data(ssv_data, format='ssv', has_header=True, columns=3)
        assert data == []

    def test_load_ssv_with_columns_skips_blank_lines(self):
        """Test columns parameter skips blank lines"""
        ssv_data = StringIO("ID Name Description\n\n1 Alice Engineer\n\n2 Bob Manager")
        data = load_data(ssv_data, format='ssv', has_header=True, columns=3)
        assert len(data) == 2
        assert data[0]['Name'] == 'Alice'
        assert data[1]['Name'] == 'Bob'

    def test_load_ssv_with_columns_and_query(self):
        """Test columns parameter works with SQL query"""
        ssv_data = StringIO("ID Name Role\n1 Alice Engineer\n2 Bob Manager\n3 Carol Engineer")
        data = load_data(
            ssv_data,
            format='ssv',
            has_header=True,
            columns=3,
            query="SELECT * FROM data WHERE Role = 'Engineer'"
        )
        assert len(data) == 2
        assert data[0]['Name'] == 'Alice'
        assert data[1]['Name'] == 'Carol'

    def test_load_spaces_alias_with_columns(self):
        """Test that 'spaces' alias works with columns parameter"""
        spaces_data = StringIO("ID Name Description\n1 Alice Software engineer\n2 Bob Project manager")
        data = load_data(spaces_data, format='spaces', has_header=True, columns=3)
        assert len(data) == 2
        assert data[0]['Description'] == 'Software engineer'
        assert data[1]['Description'] == 'Project manager'


class TestLoadJSON:
    """Tests for JSON data loading"""

    def test_load_json_file(self):
        data = load_data(str(FIXTURES_DIR / 'sample.json'), format='json')
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'
        assert data[0]['value'] == 42

    def test_load_json_inline(self):
        json_data = StringIO('[{"name": "Arthur", "value": 42}]')
        data = load_data(json_data, format='json')
        assert len(data) == 1
        assert data[0]['name'] == 'Arthur'

    def test_load_json_object(self):
        json_data = StringIO('{"title": "Test", "count": 5}')
        data = load_data(json_data, format='json')
        assert data['title'] == 'Test'
        assert data['count'] == 5


class TestLoadYAML:
    """Tests for YAML data loading"""

    def test_load_yaml_file(self):
        data = load_data(str(FIXTURES_DIR / 'sample.yaml'), format='yaml')
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'
        assert data[0]['value'] == 42

    def test_load_yaml_inline(self):
        yaml_data = StringIO("- name: Arthur\n  value: 42\n- name: Ford\n  value: 100")
        data = load_data(yaml_data, format='yaml')
        assert len(data) == 2
        assert data[0]['name'] == 'Arthur'


class TestLoadTOML:
    """Tests for TOML data loading"""

    def test_load_toml_file(self):
        data = load_data(str(FIXTURES_DIR / 'sample.toml'), format='toml')
        assert 'items' in data
        assert len(data['items']) == 3
        assert data['items'][0]['name'] == 'Arthur'
        assert data['items'][0]['value'] == 42

    def test_load_toml_inline(self):
        toml_data = StringIO('[items]\nname = "Arthur"\nvalue = 42\n\n[config]\ntitle = "Test"')
        data = load_data(toml_data, format='toml')
        assert data['items']['name'] == 'Arthur'
        assert data['items']['value'] == 42
        assert data['config']['title'] == 'Test'


class TestLoadSQLite:
    """Tests for SQLite database loading"""

    def test_load_sqlite_with_table(self):
        data = load_data(str(FIXTURES_DIR / 'sample.db'), format='sqlite', table='items')
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'
        assert data[0]['value'] == 42
        assert data[0]['category'] == 'A'

    def test_load_sqlite_with_query(self):
        data = load_data(
            str(FIXTURES_DIR / 'sample.db'),
            format='sqlite',
            query='SELECT * FROM items WHERE category = "A"'
        )
        assert len(data) == 2
        assert data[0]['name'] == 'Arthur'
        assert data[1]['name'] == 'Zaphod'

    def test_load_sqlite_query_overrides_table(self):
        """When both query and table are specified, query takes precedence"""
        data = load_data(
            str(FIXTURES_DIR / 'sample.db'),
            format='sqlite',
            table='items',
            query='SELECT * FROM metadata'
        )
        assert len(data) == 2
        assert data[0]['key'] == 'author'

    def test_load_sqlite_requires_table_or_query(self):
        """SQLite format requires either table or query parameter"""
        with pytest.raises(ValueError, match="requires either 'table' or 'query'"):
            load_data(str(FIXTURES_DIR / 'sample.db'), format='sqlite')

    def test_load_sqlite_no_inline_support(self):
        """SQLite does not support inline data"""
        with pytest.raises(ValueError, match="does not support inline data"):
            load_data(StringIO("dummy"), format='sqlite', table='items')

    def test_load_sqlite_sql_injection_prevented(self):
        """SQL injection via table name should be safely escaped"""
        # Attempt SQL injection via table name - should be escaped, not executed
        # The table doesn't exist, but the point is it doesn't execute as SQL
        from pandoc_embedz.data_loader import _quote_identifier
        # Verify quoting works correctly
        assert _quote_identifier('items') == '"items"'
        assert _quote_identifier('items; DROP TABLE items;--') == '"items; DROP TABLE items;--"'
        assert _quote_identifier('my table') == '"my table"'
        assert _quote_identifier('test"quote') == '"test""quote"'

    def test_load_sqlite_table_name_with_special_chars(self):
        """Table names with special characters should be properly quoted"""
        from pandoc_embedz.data_loader import _quote_identifier
        # These would be dangerous without quoting, but are safe when quoted
        assert _quote_identifier('123table') == '"123table"'
        assert _quote_identifier('my-table') == '"my-table"'
        assert _quote_identifier('schema.table') == '"schema.table"'


class TestLoadLines:
    """Tests for lines format data loading"""

    def test_load_lines_file(self):
        data = load_data(str(FIXTURES_DIR / 'sample.txt'), format='lines')
        assert len(data) == 3
        assert data[0] == 'Arthur'
        assert data[1] == 'Ford'
        assert data[2] == 'Zaphod'

    def test_load_lines_inline(self):
        lines_data = StringIO("Arthur\nFord\nZaphod\n")
        data = load_data(lines_data, format='lines')
        assert len(data) == 3
        assert data[0] == 'Arthur'

    def test_load_lines_skip_empty(self):
        lines_data = StringIO("Arthur\n\nFord\n\n\nZaphod")
        data = load_data(lines_data, format='lines')
        assert len(data) == 6  # Empty lines are preserved
        assert data == ['Arthur', '', 'Ford', '', '', 'Zaphod']


class TestAutoDetection:
    """Tests for format auto-detection"""

    def test_csv_auto_detect(self):
        data = load_data(str(FIXTURES_DIR / 'sample.csv'))
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'

    def test_json_auto_detect(self):
        data = load_data(str(FIXTURES_DIR / 'sample.json'))
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'

    def test_yaml_auto_detect(self):
        data = load_data(str(FIXTURES_DIR / 'sample.yaml'))
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'

    def test_toml_auto_detect(self):
        data = load_data(str(FIXTURES_DIR / 'sample.toml'))
        assert 'items' in data
        assert len(data['items']) == 3
        assert data['items'][0]['name'] == 'Arthur'

    def test_sqlite_auto_detect(self):
        """SQLite auto-detects format but still requires table parameter"""
        data = load_data(str(FIXTURES_DIR / 'sample.db'), table='items')
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'

    def test_txt_auto_detect(self):
        data = load_data(str(FIXTURES_DIR / 'sample.txt'))
        assert len(data) == 3
        assert data[0] == 'Arthur'


class TestMultiTableSQL:
    """Tests for multi-table SQL queries (integration test via process_embedz)"""

    def test_multi_table_join(self):
        """Test joining two CSV files with SQL"""
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        # Clear state
        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        # Create embedz code block with multi-table data
        code = '''---
data:
  products: tests/fixtures/products.csv
  sales: tests/fixtures/sales.csv
query: |
  SELECT
    p.product_name,
    s.quantity,
    s.date
  FROM sales s
  JOIN products p ON s.product_id = p.product_id
  ORDER BY s.date
---
{% for row in data %}
- {{ row.product_name }}: {{ row.quantity }} units on {{ row.date }}
{% endfor %}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()

        result = process_embedz(elem, doc)

        # Convert result to markdown
        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        # Verify output contains joined data
        assert 'Widget' in markdown
        assert 'Gadget' in markdown
        assert 'Doohickey' in markdown
        assert '5 units' in markdown
        assert '2024-01-15' in markdown

    def test_multi_table_aggregation(self):
        """Test aggregating data from multiple tables"""
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        # Clear state
        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        # Create embedz code block with aggregation
        code = '''---
data:
  products: tests/fixtures/products.csv
  sales: tests/fixtures/sales.csv
query: |
  SELECT
    p.product_name,
    SUM(s.quantity) as total_quantity,
    SUM(s.quantity * p.price) as total_revenue
  FROM sales s
  JOIN products p ON s.product_id = p.product_id
  GROUP BY p.product_name
  ORDER BY total_revenue DESC
---
| Product | Quantity | Revenue |
|---------|----------|---------|
{% for row in data -%}
| {{ row.product_name }} | {{ row.total_quantity }} | ${{ "%.2f" | format(row.total_revenue) }} |
{% endfor -%}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()

        result = process_embedz(elem, doc)

        # Convert result to markdown
        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        # Verify output contains aggregated data
        assert 'Widget' in markdown
        assert 'Gadget' in markdown
        assert 'Doohickey' in markdown

    def test_multi_table_without_query(self):
        """Multi-table data without query allows direct access via data.table_name"""
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        # Clear state
        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        # Create embedz code block without query - accessing via data.table_name
        code = '''---
data:
  products: tests/fixtures/products.csv
  sales: tests/fixtures/sales.csv
---
## Products
{% for p in data.products %}
- {{ p.product_name }}: ¥{{ p.price }}
{% endfor %}

## Sales
{% for s in data.sales %}
- Sale #{{ s.sale_id }}: {{ s.quantity }} units
{% endfor %}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()

        result = process_embedz(elem, doc)

        # Convert result to markdown
        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        # Verify output contains data from both files
        assert 'Products' in markdown
        assert 'Sales' in markdown
        assert 'Widget' in markdown
        assert 'Gadget' in markdown
        assert 'Sale #101' in markdown
        assert 'Sale #102' in markdown

    def test_multi_table_mixed_formats(self):
        """Multi-table can combine different formats (YAML + CSV)"""
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        # Clear state
        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        # Create embedz code block combining YAML config and CSV data
        code = '''---
data:
  config: tests/fixtures/config.yaml
  sales: tests/fixtures/sales.csv
---
# {{ data.config.title }}
## {{ data.config.subtitle }}

By {{ data.config.author }} (v{{ data.config.version }})

{% for sale in data.sales[:3] %}
- Sale #{{ sale.sale_id }}: {{ sale.quantity }} units
{% endfor %}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()

        result = process_embedz(elem, doc)

        # Convert result to markdown
        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        # Verify output contains both config and data
        assert '2024 Sales Report' in markdown
        assert 'Q1 Results' in markdown
        assert 'John Doe' in markdown
        assert 'v1.0' in markdown
        assert 'Sale #101' in markdown

    def test_multi_table_inline_csv(self):
        """Multi-table with inline CSV data"""
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        # Clear state
        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        # Create embedz code block with inline CSV data
        code = '''---
data:
  products: |
    product_id,product_name,price
    1,Widget,1280
    2,Gadget,2480
  sales: |
    sale_id,product_id,quantity
    101,1,5
    102,2,3
---
## Products
{% for p in data.products %}
- {{ p.product_name }}: ¥{{ "{:,}".format(p.price|int) }}
{% endfor %}

## Sales
{% for s in data.sales %}
- Sale #{{ s.sale_id }}: {{ s.quantity }} units
{% endfor %}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()

        result = process_embedz(elem, doc)

        # Convert result to markdown
        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        # Verify output contains data from inline sources
        assert 'Products' in markdown
        assert 'Widget: ¥1,280' in markdown
        assert 'Gadget: ¥2,480' in markdown
        assert 'Sale #101: 5 units' in markdown
        assert 'Sale #102: 3 units' in markdown

    def test_multi_table_inline_yaml(self):
        """Multi-table with inline YAML config and CSV data"""
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        # Clear state
        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        # Create embedz code block with mixed inline data
        code = '''---
data:
  config:
    format: yaml
    data: |
      title: "Test Report"
      year: 2024
  sales: |
    date,amount
    2024-01-01,100
    2024-01-02,200
---
# {{ data.config.title }} ({{ data.config.year }})

{% for s in data.sales %}
- {{ s.date }}: ${{ s.amount }}
{% endfor %}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()

        result = process_embedz(elem, doc)

        # Convert result to markdown
        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        # Verify output contains both config and data
        assert 'Test Report (2024)' in markdown
        assert '2024-01-01: \\$100' in markdown or '2024-01-01: $100' in markdown
        assert '2024-01-02: \\$200' in markdown or '2024-01-02: $200' in markdown

    def test_multi_table_mixed_inline_and_file(self):
        """Multi-table with both inline data and file paths"""
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        # Clear state
        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        # Create embedz code block mixing inline and file data
        code = '''---
data:
  config:
    format: yaml
    data: |
      title: "Mixed Source Report"
  products: tests/fixtures/products.csv
  sales: |
    sale_id,product_id,quantity
    999,1,10
---
# {{ data.config.title }}

## Products from file
{% for p in data.products[:2] %}
- {{ p.product_name }}
{% endfor %}

## Sales from inline
{% for s in data.sales %}
- Sale #{{ s.sale_id }}: {{ s.quantity }} units
{% endfor %}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()

        result = process_embedz(elem, doc)

        # Convert result to markdown
        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        # Verify output contains data from both sources
        assert 'Mixed Source Report' in markdown
        assert 'Widget' in markdown  # From file
        assert 'Sale #999: 10 units' in markdown  # From inline

    def test_data_file_and_data_part_mutually_exclusive(self):
        """Error should be raised if both data attribute and inline data are specified"""
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        # Clear state
        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        # Create embedz code block with both data attribute and inline data
        code = '''---
data: tests/fixtures/products.csv
---
{% for p in data %}
- {{ p.product_name }}
{% endfor %}
---
product_id,product_name,price
1,Widget,1280
2,Gadget,2480'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()

        # Should raise ValueError
        import pytest
        with pytest.raises(ValueError, match="Cannot specify both 'data' attribute and inline data"):
            process_embedz(elem, doc)

class TestEmptyInput:
    """Tests for empty input handling"""

    def test_empty_json_returns_empty_list(self):
        """Empty JSON input should return empty list instead of error"""
        data = load_data(StringIO(""), format='json')
        assert data == []

    def test_whitespace_json_returns_empty_list(self):
        """Whitespace-only JSON input should return empty list"""
        data = load_data(StringIO("   \n  \t  "), format='json')
        assert data == []

    def test_empty_csv_returns_empty_list(self):
        """Empty CSV input should return empty list instead of error"""
        data = load_data(StringIO(""), format='csv')
        assert data == []

    def test_whitespace_csv_returns_empty_list(self):
        """Whitespace-only CSV input should return empty list"""
        data = load_data(StringIO("   \n  \t  "), format='csv')
        assert data == []

    def test_empty_tsv_returns_empty_list(self):
        """Empty TSV input should return empty list"""
        data = load_data(StringIO(""), format='tsv')
        assert data == []


class TestLoadExcel:
    """Tests for Excel data loading"""

    @pytest.fixture(autouse=True)
    def check_openpyxl(self):
        pytest.importorskip("openpyxl")

    def test_excel_extensions(self):
        """Extension auto-detection for .xlsx and .xls"""
        assert guess_format_from_filename('data.xlsx') == 'excel'
        assert guess_format_from_filename('data.xls') == 'excel'

    def test_load_excel_with_header(self):
        """Load Excel with header row (default)"""
        data = load_data(str(FIXTURES_DIR / 'sample.xlsx'), format='excel', table='items')
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'
        assert data[0]['value'] == 42
        assert data[0]['category'] == 'A'

    def test_load_excel_without_header(self):
        """Load Excel without header row"""
        data = load_data(str(FIXTURES_DIR / 'sample.xlsx'), format='excel', table='items', has_header=False)
        assert len(data) == 4  # Including header row as data
        assert isinstance(data[0], list)
        assert data[0][0] == 'name'  # Header row treated as data

    def test_load_excel_with_sheet_name(self):
        """Load specific sheet by name via table parameter"""
        data = load_data(str(FIXTURES_DIR / 'sample.xlsx'), format='excel', table='metadata')
        assert len(data) == 2
        assert data[0]['key'] == 'author'
        assert data[0]['description'] == 'Douglas Adams'

    def test_load_excel_default_sheet(self):
        """Load first sheet when table is not specified"""
        data = load_data(str(FIXTURES_DIR / 'sample.xlsx'), format='excel')
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'

    def test_load_excel_with_query(self):
        """Apply SQL query to Excel data"""
        data = load_data(
            str(FIXTURES_DIR / 'sample.xlsx'),
            format='excel',
            table='items',
            query='SELECT * FROM data WHERE category = "A"'
        )
        assert len(data) == 2
        assert data[0]['name'] == 'Arthur'
        assert data[1]['name'] == 'Zaphod'

    def test_load_excel_no_inline(self):
        """Excel does not support inline data"""
        with pytest.raises(ValueError, match="does not support inline data"):
            load_data(StringIO("dummy"), format='excel')

    def test_load_excel_no_openpyxl(self, monkeypatch):
        """Error message when openpyxl is not installed"""
        import pandoc_embedz.data_loader as dl
        monkeypatch.setattr(dl, 'openpyxl', None)
        with pytest.raises(ImportError, match="openpyxl"):
            load_data(str(FIXTURES_DIR / 'sample.xlsx'), format='excel')

    def test_load_excel_skips_blank_rows_and_columns(self, tmp_path):
        """Leading blank rows and all-blank columns are skipped"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append([])                              # blank row
        ws.append([])                              # blank row
        ws.append([None, 'name', 'value', None])   # header with blank cols
        ws.append([None, 'Arthur', 42, None])
        ws.append([None, 'Ford', 100, None])
        path = str(tmp_path / 'blank.xlsx')
        wb.save(path)

        data = load_data(path, format='excel')
        assert len(data) == 2
        assert data[0] == {'name': 'Arthur', 'value': 42}
        assert data[1] == {'name': 'Ford', 'value': 100}

    def test_load_excel_skips_blank_rows_without_header(self, tmp_path):
        """Blank rows/columns skipped in headerless mode too"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append([])
        ws.append([None, 'Arthur', 42, None])
        ws.append([None, 'Ford', 100, None])
        path = str(tmp_path / 'blank_noheader.xlsx')
        wb.save(path)

        data = load_data(path, format='excel', has_header=False)
        assert len(data) == 2
        assert data[0] == ['Arthur', 42]
        assert data[1] == ['Ford', 100]

    def test_load_excel_empty_sheet(self, tmp_path, capsys):
        """Empty sheet returns empty list with warning"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        # Don't add any data
        path = str(tmp_path / 'empty.xlsx')
        wb.save(path)

        assert load_data(path, format='excel') == []
        assert 'contains no data' in capsys.readouterr().err

        assert load_data(path, format='excel', has_header=False) == []

    def test_load_excel_transpose_with_header(self, tmp_path):
        """Transpose swaps rows and columns, first column becomes header"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['name', 'Arthur', 'Ford', 'Zaphod'])
        ws.append(['value', 42, 100, 99])
        ws.append(['category', 'A', 'B', 'A'])
        path = str(tmp_path / 'transposed.xlsx')
        wb.save(path)

        data = load_data(path, format='excel', transpose=True)
        assert len(data) == 3
        assert data[0] == {'name': 'Arthur', 'value': 42, 'category': 'A'}
        assert data[1] == {'name': 'Ford', 'value': 100, 'category': 'B'}
        assert data[2] == {'name': 'Zaphod', 'value': 99, 'category': 'A'}

    def test_load_excel_transpose_without_header(self, tmp_path):
        """Transpose without header returns list of lists"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['Arthur', 'Ford'])
        ws.append([42, 100])
        path = str(tmp_path / 'transposed_noheader.xlsx')
        wb.save(path)

        data = load_data(path, format='excel', transpose=True, has_header=False)
        assert len(data) == 2
        assert data[0] == ['Arthur', 42]
        assert data[1] == ['Ford', 100]

    def test_load_excel_transpose_with_query(self, tmp_path):
        """Transpose works with SQL query"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['name', 'Arthur', 'Ford', 'Zaphod'])
        ws.append(['value', 42, 100, 99])
        path = str(tmp_path / 'transposed_query.xlsx')
        wb.save(path)

        data = load_data(
            path, format='excel', transpose=True,
            query='SELECT * FROM data WHERE value >= 99'
        )
        assert len(data) == 2
        assert data[0]['name'] == 'Ford'
        assert data[1]['name'] == 'Zaphod'

    def test_load_excel_skiprows(self, tmp_path):
        """Skip leading description rows"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['Annual Report 2024'])
        ws.append(['Generated: 2024-04-01'])
        ws.append([])
        ws.append(['name', 'value'])
        ws.append(['Arthur', 42])
        ws.append(['Ford', 100])
        path = str(tmp_path / 'with_title.xlsx')
        wb.save(path)

        data = load_data(path, format='excel', skiprows=2)
        assert len(data) == 2
        assert data[0] == {'name': 'Arthur', 'value': 42}
        assert data[1] == {'name': 'Ford', 'value': 100}

    def test_load_excel_skiprows_pattern_any_column(self, tmp_path):
        """Skip rows by pattern matching any column"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['Title'])
        ws.append(['Description'])
        ws.append([])
        ws.append(['name', 'value'])
        ws.append(['Arthur', 42])
        path = str(tmp_path / 'pattern.xlsx')
        wb.save(path)

        data = load_data(path, format='excel', skiprows='name')
        assert len(data) == 1
        assert data[0] == {'name': 'Arthur', 'value': 42}

    def test_load_excel_skiprows_pattern_specific_column(self, tmp_path):
        """Skip rows by pattern matching a specific column"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['info', 'Title'])
        ws.append([])
        ws.append(['name', 'value'])
        ws.append(['Arthur', 42])
        path = str(tmp_path / 'col_pattern.xlsx')
        wb.save(path)

        # Column 2 contains "value" in the header row
        data = load_data(path, format='excel', skiprows='2:value')
        assert len(data) == 1
        assert data[0] == {'name': 'Arthur', 'value': 42}

    def test_load_excel_skiprows_pattern_list(self, tmp_path):
        """Skip rows by list of patterns (all must match)"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['Title', 'Report'])
        ws.append(['年', '月', '件数'])
        ws.append([2024, 4, 100])
        ws.append([2024, 5, 200])
        path = str(tmp_path / 'list_pattern.xlsx')
        wb.save(path)

        data = load_data(path, format='excel', skiprows=['年', '月'])
        assert len(data) == 2
        assert data[0] == {'年': 2024, '月': 4, '件数': 100}
        assert data[1] == {'年': 2024, '月': 5, '件数': 200}

    def test_load_excel_skiprows_pattern_list_with_column(self, tmp_path):
        """Skip rows by list with N:text format"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['Header', 'Info'])
        ws.append(['年', '月', '件数'])
        ws.append([2024, 4, 100])
        path = str(tmp_path / 'list_col.xlsx')
        wb.save(path)

        data = load_data(path, format='excel', skiprows=['1:年', '2:月'])
        assert len(data) == 1
        assert data[0] == {'年': 2024, '月': 4, '件数': 100}

    def test_load_excel_skiprows_pattern_list_not_found(self, tmp_path):
        """Error when not all list patterns found in same row"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['年', '件数'])
        ws.append([2024, 100])
        path = str(tmp_path / 'list_notfound.xlsx')
        wb.save(path)

        with pytest.raises(ValueError, match="not found"):
            load_data(path, format='excel', skiprows=['年', '月'])

    def test_load_excel_skiprows_single_element_list(self, tmp_path):
        """Single-element list behaves like a string pattern"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['Title'])
        ws.append(['name', 'value'])
        ws.append(['Arthur', 42])
        path = str(tmp_path / 'single_list.xlsx')
        wb.save(path)

        data = load_data(path, format='excel', skiprows=['name'])
        assert len(data) == 1
        assert data[0] == {'name': 'Arthur', 'value': 42}

    def test_load_excel_skiprows_pattern_not_found(self, tmp_path):
        """Error when skiprows pattern not found"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['name', 'value'])
        ws.append(['Arthur', 42])
        path = str(tmp_path / 'no_match.xlsx')
        wb.save(path)

        with pytest.raises(ValueError, match="not found"):
            load_data(path, format='excel', skiprows='nonexistent')

    def test_load_excel_empty_header_column(self, tmp_path):
        """Empty first header cell gets replaced with column_N"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append([None, '4月', '5月', '6月'])
        ws.append(['件数', 10, 20, 30])
        ws.append(['割合', 0.1, 0.2, 0.3])
        path = str(tmp_path / 'empty_header.xlsx')
        wb.save(path)

        data = load_data(path, format='excel')
        assert len(data) == 2
        assert data[0]['column_0'] == '件数'
        assert data[0]['4月'] == 10
        assert data[1]['column_0'] == '割合'
        assert data[1]['6月'] == 0.3

    def test_load_excel_duplicate_column_names(self, tmp_path):
        """Duplicate column names get numeric suffix"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append(['name', 'value', 'value'])
        ws.append(['Arthur', 42, 100])
        path = str(tmp_path / 'dup_cols.xlsx')
        wb.save(path)

        data = load_data(path, format='excel')
        assert len(data) == 1
        assert data[0]['name'] == 'Arthur'
        assert data[0]['value'] == 42
        assert data[0]['value_1'] == 100

    def test_load_excel_whitespace_column_names(self, tmp_path):
        """Column names with leading/trailing whitespace are stripped"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append([' name ', ' value '])
        ws.append(['Arthur', 42])
        path = str(tmp_path / 'ws_cols.xlsx')
        wb.save(path)

        data = load_data(path, format='excel')
        assert data[0]['name'] == 'Arthur'
        assert data[0]['value'] == 42

    def test_load_excel_nan_in_data_without_header(self, tmp_path):
        """NaN values in headerless data become empty strings"""
        import openpyxl
        wb = openpyxl.Workbook()
        ws = wb.active
        ws.append([None, '4月', '5月'])
        ws.append(['件数', 10, None])
        path = str(tmp_path / 'nan_data.xlsx')
        wb.save(path)

        data = load_data(path, format='excel', has_header=False)
        assert data[0] == ['', '4月', '5月']
        assert data[1] == ['件数', 10, '']

    def test_excel_auto_detect(self):
        """Format auto-detection for .xlsx files"""
        data = load_data(str(FIXTURES_DIR / 'sample.xlsx'))
        assert len(data) == 3
        assert data[0]['name'] == 'Arthur'


class TestNormalizeDataSource:
    """Tests for _normalize_data_source 3-tuple return"""

    def test_string_path_returns_3_tuple(self):
        """String path returns (source, format, {})"""
        from pandoc_embedz.data_loader import _normalize_data_source
        source, fmt, kwargs = _normalize_data_source('data.csv', 'test')
        assert source == 'data.csv'
        assert fmt == 'csv'
        assert kwargs == {}

    def test_inline_data_returns_3_tuple(self):
        """Inline data dict returns (StringIO, format, {})"""
        from pandoc_embedz.data_loader import _normalize_data_source
        source, fmt, kwargs = _normalize_data_source(
            {'data': 'a,b\n1,2'}, 'test'
        )
        assert isinstance(source, StringIO)
        assert fmt == 'csv'
        assert kwargs == {}

    def test_file_dict_returns_3_tuple(self):
        """file: dict returns (path, format, extra_kwargs)"""
        from pandoc_embedz.data_loader import _normalize_data_source
        source, fmt, kwargs = _normalize_data_source(
            {'file': 'data/test.xlsx', 'table': 'Sheet1', 'skiprows': 'Year'},
            'test'
        )
        assert source == 'data/test.xlsx'
        assert fmt == 'excel'
        assert kwargs == {'table': 'Sheet1', 'skiprows': 'Year'}

    def test_file_dict_with_explicit_format(self):
        """file: dict with explicit format overrides auto-detection"""
        from pandoc_embedz.data_loader import _normalize_data_source
        source, fmt, kwargs = _normalize_data_source(
            {'file': 'data.dat', 'format': 'tsv'},
            'test'
        )
        assert fmt == 'tsv'
        assert kwargs == {}

    def test_file_dict_with_data_format_fallback(self):
        """file: dict uses data_format when no format in dict or extension"""
        from pandoc_embedz.data_loader import _normalize_data_source
        source, fmt, kwargs = _normalize_data_source(
            {'file': 'data.dat'},
            'test',
            data_format='json'
        )
        assert fmt == 'json'

    def test_dict_without_data_or_file_raises(self):
        """Dict without 'data' or 'file' key raises ValueError"""
        from pandoc_embedz.data_loader import _normalize_data_source
        with pytest.raises(ValueError, match="must have 'data' or 'file' key"):
            _normalize_data_source({'table': 'Sheet1'}, 'test')

    def test_multiline_string_returns_3_tuple(self):
        """Multiline string returns (StringIO, csv, {})"""
        from pandoc_embedz.data_loader import _normalize_data_source
        source, fmt, kwargs = _normalize_data_source('a,b\n1,2', 'test')
        assert isinstance(source, StringIO)
        assert fmt == 'csv'
        assert kwargs == {}


class TestIsResolvedData:
    """Tests for _is_resolved_data with file: dict"""

    def test_list_is_resolved(self):
        from pandoc_embedz.data_loader import _is_resolved_data
        assert _is_resolved_data([1, 2, 3]) is True

    def test_plain_dict_is_resolved(self):
        from pandoc_embedz.data_loader import _is_resolved_data
        assert _is_resolved_data({'key': 'value'}) is True

    def test_inline_data_dict_is_not_resolved(self):
        from pandoc_embedz.data_loader import _is_resolved_data
        assert _is_resolved_data({'data': 'a,b\n1,2'}) is False

    def test_format_dict_is_not_resolved(self):
        from pandoc_embedz.data_loader import _is_resolved_data
        assert _is_resolved_data({'format': 'csv'}) is False

    def test_file_dict_is_not_resolved(self):
        from pandoc_embedz.data_loader import _is_resolved_data
        assert _is_resolved_data({'file': 'data.xlsx', 'table': 'Sheet1'}) is False


class TestMultiTableExcel:
    """Tests for multi-table SQL with Excel files"""

    @pytest.fixture(autouse=True)
    def check_openpyxl(self):
        pytest.importorskip("openpyxl")

    def test_multi_table_excel_with_query(self, tmp_path):
        """Multi-table SQL query with Excel data sources"""
        import openpyxl
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        # Create Excel file with two sheets
        wb = openpyxl.Workbook()
        ws1 = wb.active
        ws1.title = 'products'
        ws1.append(['product_id', 'name', 'price'])
        ws1.append([1, 'Widget', 1280])
        ws1.append([2, 'Gadget', 2480])

        ws2 = wb.create_sheet('sales')
        ws2.append(['product_id', 'quantity'])
        ws2.append([1, 5])
        ws2.append([2, 3])

        path = str(tmp_path / 'multi.xlsx')
        wb.save(path)

        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        code = f'''---
data:
  products:
    file: {path}
    table: products
  sales:
    file: {path}
    table: sales
query: |
  SELECT p.name, p.price, s.quantity
  FROM sales s
  JOIN products p ON s.product_id = p.product_id
  ORDER BY p.name
---
{{% for row in data %}}
- {{{{ row.name }}}}: {{{{ row.quantity }}}} units at ¥{{{{ row.price }}}}
{{% endfor %}}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()
        result = process_embedz(elem, doc)

        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        assert 'Gadget' in markdown
        assert 'Widget' in markdown
        assert '5 units' in markdown
        assert '3 units' in markdown

    def test_multi_table_excel_with_skiprows(self, tmp_path):
        """Multi-table SQL with skiprows parameter in file: dict"""
        import openpyxl
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        wb = openpyxl.Workbook()
        ws = wb.active
        ws.title = 'data'
        ws.append(['Report Title'])
        ws.append(['Generated 2024'])
        ws.append([])
        ws.append(['name', 'value'])
        ws.append(['Arthur', 42])
        ws.append(['Ford', 100])

        path = str(tmp_path / 'skiprows.xlsx')
        wb.save(path)

        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        code = f'''---
data:
  items:
    file: {path}
    table: data
    skiprows: name
query: |
  SELECT * FROM items ORDER BY value DESC
---
{{% for row in data %}}
- {{{{ row.name }}}}: {{{{ row.value }}}}
{{% endfor %}}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()
        result = process_embedz(elem, doc)

        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        assert 'Ford' in markdown
        assert 'Arthur' in markdown

    def test_multi_table_file_dict_without_query(self, tmp_path):
        """Multi-table with file: dict but no query (direct access)"""
        import openpyxl
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        wb = openpyxl.Workbook()
        ws1 = wb.active
        ws1.title = 'Sheet1'
        ws1.append(['name', 'value'])
        ws1.append(['Arthur', 42])
        ws2 = wb.create_sheet('Sheet2')
        ws2.append(['key', 'desc'])
        ws2.append(['author', 'Douglas Adams'])

        path = str(tmp_path / 'direct.xlsx')
        wb.save(path)

        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        code = f'''---
data:
  products:
    file: {path}
    table: Sheet1
  meta:
    file: {path}
    table: Sheet2
---
{{% for p in data.products %}}
- {{{{ p.name }}}}: {{{{ p.value }}}}
{{% endfor %}}
Author: {{{{ data.meta[0].desc }}}}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()
        result = process_embedz(elem, doc)

        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        assert 'Arthur: 42' in markdown
        assert 'Douglas Adams' in markdown

    def test_multi_table_mixed_file_dict_and_string(self, tmp_path):
        """Mix file: dict (Excel) with plain string (CSV)"""
        import openpyxl
        from pandoc_embedz.filter import process_embedz, GLOBAL_VARS
        from pandoc_embedz.config import SAVED_TEMPLATES
        import panflute as pf

        wb = openpyxl.Workbook()
        ws = wb.active
        ws.title = 'items'
        ws.append(['id', 'name'])
        ws.append([1, 'Widget'])
        ws.append([2, 'Gadget'])

        xlsx_path = str(tmp_path / 'items.xlsx')
        wb.save(xlsx_path)

        csv_path = str(tmp_path / 'prices.csv')
        Path(csv_path).write_text('id,price\n1,1280\n2,2480\n')

        SAVED_TEMPLATES.clear()
        GLOBAL_VARS.clear()

        code = f'''---
data:
  items:
    file: {xlsx_path}
    table: items
  prices: {csv_path}
query: |
  SELECT i.name, p.price
  FROM items i JOIN prices p ON i.id = p.id
  ORDER BY i.name
---
{{% for row in data %}}
- {{{{ row.name }}}}: ¥{{{{ row.price }}}}
{{% endfor %}}'''

        elem = pf.CodeBlock(code, classes=['embedz'])
        doc = pf.Doc()
        result = process_embedz(elem, doc)

        if isinstance(result, list):
            markdown = pf.convert_text(result, input_format='panflute', output_format='markdown')
        else:
            markdown = pf.convert_text([result], input_format='panflute', output_format='markdown')

        assert 'Gadget' in markdown
        assert 'Widget' in markdown
        assert '1280' in markdown or '1,280' in markdown



"""Data loading module for pandoc-embedz

Handles loading data from various formats using a dispatch table pattern.
"""

from typing import Dict, Any, List, Optional, Union, Tuple
from functools import partial
import pandas as pd
import yaml
import json
import sqlite3
import sys
from io import StringIO
from pathlib import Path

try:
    import tomllib
except ImportError:
    try:
        import tomli as tomllib
    except ImportError:
        tomllib = None

try:
    from sqlite_utils import Database as SqliteUtilsDatabase
except ImportError:
    SqliteUtilsDatabase = None

try:
    import openpyxl  # noqa: F401 - imported for availability check
except ImportError:
    openpyxl = None

# ─────────────────────────────────────────────────────────────────────────────
# Constants

FORMAT_EXTENSIONS = {
    '.txt': 'lines',
    '.tsv': 'tsv',
    '.json': 'json',
    '.yaml': 'yaml',
    '.yml': 'yaml',
    '.toml': 'toml',
    '.db': 'sqlite',
    '.sqlite': 'sqlite',
    '.sqlite3': 'sqlite',
    '.xlsx': 'excel',
    '.xls': 'excel',
}

DEFAULT_FORMAT = 'csv'

# ─────────────────────────────────────────────────────────────────────────────
# SQL Query Support

def _apply_sql_query(df: pd.DataFrame, query: str, table_name: str = 'data') -> List[Dict[str, Any]]:
    """Apply SQL query to DataFrame using in-memory SQLite"""
    return _apply_sql_query_multi({table_name: df}, query)

def _apply_sql_query_multi(tables: Dict[str, pd.DataFrame], query: str) -> List[Dict[str, Any]]:
    """Apply SQL query to multiple DataFrames using in-memory SQLite"""
    conn = sqlite3.connect(':memory:')
    conn.row_factory = sqlite3.Row

    try:
        for table_name, df in tables.items():
            df.to_sql(table_name, conn, index=False, if_exists='replace')

        cursor = conn.cursor()
        cursor.execute(query)
        rows = cursor.fetchall()

        return [dict(row) for row in rows]
    finally:
        conn.close()

# ─────────────────────────────────────────────────────────────────────────────
# Format Loaders

def _build_csv_read_kwargs(sep: str) -> Dict[str, Any]:
    r"""Build pandas read_csv kwargs for the given separator.

    Args:
        sep: Separator string (e.g., ',', '\t', r'\s+')

    Returns:
        Dict of kwargs for pd.read_csv()
    """
    kwargs: Dict[str, Any] = {'sep': sep}

    if sep == r'\s+':
        kwargs['engine'] = 'python'

    return kwargs

def _read_source(source: Union[str, StringIO], comment: str = 'none') -> str:
    """Read content from file path or StringIO.

    Args:
        source: File path string or StringIO object
        comment: Comment handling mode:
            'none' - no comment handling (default)
            'line' - strip all '#' lines
            'head' - strip leading '#' lines only

    Returns:
        Content as string
    """
    content = source.getvalue() if isinstance(source, StringIO) else Path(source).read_text(encoding='utf-8')
    if comment == 'head':
        lines = content.splitlines()
        while lines and lines[0].startswith('#'):
            lines.pop(0)
        content = '\n'.join(lines)
    elif comment == 'line':
        content = '\n'.join(
            line for line in content.splitlines()
            if not line.startswith('#')
        )
    return content

def _load_json(source: Union[str, StringIO], **kwargs) -> Union[List[Any], Dict[str, Any]]:
    """Load JSON format

    Returns empty list for empty input instead of raising an error.
    """
    content = _read_source(source, comment=kwargs.pop('comment', 'none'))

    # Handle empty input - return empty list
    if not content.strip():
        return []

    return json.loads(content)

def _load_yaml(source: Union[str, StringIO], **kwargs) -> Union[List[Any], Dict[str, Any]]:
    """Load YAML format"""
    content = _read_source(source, comment=kwargs.pop('comment', 'none'))
    return yaml.safe_load(content)

def _load_toml(source: Union[str, StringIO], **kwargs) -> Dict[str, Any]:
    """Load TOML format"""
    if tomllib is None:
        raise ImportError(
            "TOML support requires 'tomli' package for Python < 3.11. "
            "Install with: pip install tomli"
        )
    content = _read_source(source, comment=kwargs.pop('comment', 'none'))
    return tomllib.loads(content)

def _quote_identifier(name: str) -> str:
    """Quote SQLite identifier (table/column name) to prevent SQL injection.

    Uses double-quote escaping per SQL standard (supported by SQLite).
    Any double quotes in the name are escaped by doubling them.

    Args:
        name: Identifier name to quote

    Returns:
        str: Safely quoted identifier

    Examples:
        >>> _quote_identifier('items')
        '"items"'
        >>> _quote_identifier('my table')
        '"my table"'
        >>> _quote_identifier('test"quote')
        '"test""quote"'
    """
    return '"' + name.replace('"', '""') + '"'


def _load_sqlite(source: Union[str, StringIO], **kwargs) -> List[Dict[str, Any]]:
    """Load SQLite database

    Uses sqlite-utils if available for improved API, falls back to sqlite3.
    """
    if isinstance(source, StringIO):
        raise ValueError(
            "SQLite format does not support inline data. "
            "Use an external .db/.sqlite/.sqlite3 file."
        )

    query = kwargs.get('query')
    table = kwargs.get('table')

    if not query and not table:
        raise ValueError("SQLite format requires either 'table' or 'query' parameter")

    # Use sqlite-utils if available (cleaner API)
    if SqliteUtilsDatabase is not None:
        db = SqliteUtilsDatabase(source)
        if query:
            # Execute query and convert to dicts using column names
            cursor = db.execute(query)
            columns = [desc[0] for desc in cursor.description]
            return [dict(zip(columns, row)) for row in cursor.fetchall()]
        else:
            return list(db[table].rows)

    # Fallback to standard sqlite3
    conn = sqlite3.connect(source)
    conn.row_factory = sqlite3.Row

    try:
        cursor = conn.cursor()
        if query:
            cursor.execute(query)
        else:
            # Quote table name to prevent SQL injection
            cursor.execute(f"SELECT * FROM {_quote_identifier(table)}")

        rows = cursor.fetchall()
        return [dict(row) for row in rows]
    finally:
        conn.close()

def _clean_column_names(columns: list) -> list:
    """Clean up column names: strip whitespace, replace empty/NaN, deduplicate.

    Follows the same pattern as script/xls2sqlite:
    - Empty or NaN column names are replaced with 'column_N' (0-based index)
    - String column names are stripped of leading/trailing whitespace
    - Duplicate names get a numeric suffix: 'name_1', 'name_2', etc.
    """
    cleaned = []
    seen: Dict[str, int] = {}
    for i, col in enumerate(columns):
        if pd.isna(col):
            col = f'column_{i}'
        elif isinstance(col, str):
            col = col.strip()
            if col == '':
                col = f'column_{i}'
        else:
            col = str(col)

        if col in seen:
            seen[col] += 1
            col = f'{col}_{seen[col]}'
        else:
            seen[col] = 0

        cleaned.append(col)
    return cleaned


def _parse_skip_pattern(pattern: str) -> Tuple[Optional[int], str]:
    """Parse a skiprows pattern string into (column_index, search_text).

    Args:
        pattern: Pattern string - either "text" or "N:text" (N is 1-based)

    Returns:
        Tuple of (column_index, search_text).
        column_index is None for "text" format, or 0-based int for "N:text".
    """
    if ':' in pattern:
        prefix, rest = pattern.split(':', 1)
        try:
            return (int(prefix) - 1, rest)
        except ValueError:
            pass  # Not "N:text" format, use entire string
    return (None, pattern)


def _row_matches_pattern(
    df: pd.DataFrame, row_idx: int, col_index: Optional[int], search_text: str
) -> bool:
    """Check if a row contains a cell exactly matching the given text.

    Args:
        df: DataFrame to search
        row_idx: Row index to check
        col_index: Column index (0-based) to check, or None to check all columns
        search_text: Text to match (exact match via str comparison)

    Returns:
        True if matching cell found, False otherwise.
    """
    cols = [col_index] if col_index is not None else range(len(df.columns))
    for j in cols:
        if j < len(df.columns):
            cell = df.iloc[row_idx, j]
            if pd.notna(cell) and str(cell) == search_text:
                return True
    return False


def _skip_to_matching_row(
    df: pd.DataFrame, pattern: Union[str, List[str]], source: str, table: Optional[str]
) -> pd.DataFrame:
    """Skip rows until finding one matching the pattern(s).

    Args:
        df: DataFrame to search
        pattern: Pattern string, or list of pattern strings (AND logic).
            Each pattern can be "text" or "N:text" (N is 1-based column index).
        source: File path (for error messages)
        table: Sheet name (for error messages), or None

    Returns:
        DataFrame starting from the matching row.

    Raises:
        ValueError: If no matching row is found.
    """
    patterns = pattern if isinstance(pattern, list) else [pattern]
    parsed = [_parse_skip_pattern(p) for p in patterns]
    for i in range(len(df)):
        if all(_row_matches_pattern(df, i, col_idx, text) for col_idx, text in parsed):
            return df.iloc[i:].reset_index(drop=True)

    sheet_info = f" (sheet: {table})" if table else ""
    raise ValueError(
        f"skiprows pattern '{pattern}' not found in "
        f"Excel file '{source}'{sheet_info}"
    )


def _load_excel(
    source: Union[str, StringIO],
    has_header: bool = True,
    **kwargs
) -> Union[List[Dict[str, Any]], List[List[Any]]]:
    """Load Excel file (.xlsx, .xls)

    Requires openpyxl package for .xlsx files.

    Args:
        source: File path to Excel file
        has_header: Whether first row is header
        **kwargs: Options including 'table' (sheet name), 'query' (SQL),
                  'transpose' (swap rows/columns), and 'skiprows' (rows to skip)

    Returns:
        List of dicts (with header) or list of lists (without header)
    """
    if isinstance(source, StringIO):
        raise ValueError(
            "Excel format does not support inline data. "
            "Use an external .xlsx/.xls file."
        )

    if openpyxl is None:
        raise ImportError(
            "Excel support requires 'openpyxl' package. "
            "Install with: pip install openpyxl"
        )

    table = kwargs.get('table')
    sheet_name = table if table is not None else 0

    # Read sheet; skip leading rows if requested
    skiprows = kwargs.get('skiprows')
    read_kwargs: Dict[str, Any] = {'sheet_name': sheet_name, 'header': None}
    if isinstance(skiprows, int):
        read_kwargs['skiprows'] = skiprows
    df = pd.read_excel(source, **read_kwargs)
    if isinstance(skiprows, (str, list)):
        df = _skip_to_matching_row(df, skiprows, source, table)

    # Drop rows and columns where all values are NaN
    df = df.dropna(how='all').dropna(axis=1, how='all')
    df = df.reset_index(drop=True)

    if df.empty:
        sheet_info = f" (sheet: {table})" if table else ""
        sys.stderr.write(
            f"Warning: Excel file '{source}'{sheet_info} contains no data\n"
        )
        return []

    # Transpose if requested (swap rows and columns)
    if kwargs.get('transpose'):
        df = df.T.reset_index(drop=True)
        df.columns = range(len(df.columns))

    # Replace NaN with empty string for template-friendly output
    df = df.astype(object).where(df.notna(), '')

    if has_header:
        # Use first row as header, clean up names
        df.columns = _clean_column_names(df.iloc[0].tolist())
        df = df.iloc[1:].reset_index(drop=True)
        if 'query' in kwargs:
            return _apply_sql_query(df, kwargs['query'])
        return df.to_dict('records')
    else:
        return df.values.tolist()


def _load_lines(source: Union[str, StringIO], **kwargs) -> List[str]:
    """Load plain text lines

    Returns each line as a string. Empty lines are preserved as empty strings.
    """
    content = _read_source(source, comment=kwargs.pop('comment', 'none'))
    return content.splitlines()

def _load_ssv_with_columns(
    source: Union[str, StringIO],
    columns: int,
    has_header: bool = True,
    **kwargs
) -> Union[List[Dict[str, Any]], List[List[Any]]]:
    """Load SSV format with fixed column count.

    Uses str.split(maxsplit=columns-1) to preserve spaces in the last column.

    Args:
        source: File path or StringIO object
        columns: Number of columns (last column gets all remaining content)
        has_header: Whether first line is header
        **kwargs: Additional options (e.g., query)

    Returns:
        List of dicts (with header) or list of lists (without header)
    """
    comment_mode = kwargs.pop('comment', 'line')
    if comment_mode is False:
        comment_mode = 'none'
    content = _read_source(source, comment=comment_mode)
    lines = content.splitlines()

    if not lines:
        return []

    maxsplit = columns - 1
    result: List[List[str]] = []
    header: Optional[List[str]] = None

    for line in lines:
        if not line.strip():
            continue

        parts = line.split(maxsplit=maxsplit)

        # Pad with empty strings if fewer columns than expected
        while len(parts) < columns:
            parts.append('')

        if has_header and header is None:
            header = parts
        else:
            result.append(parts)

    if has_header and header is not None:
        records = [dict(zip(header, row)) for row in result]
        if 'query' in kwargs:
            df = pd.DataFrame(records)
            return _apply_sql_query(df, kwargs['query'])
        return records
    else:
        return result


def _load_csv(
    source: Union[str, StringIO],
    sep: str = ',',
    has_header: bool = True,
    **kwargs
) -> Union[List[Dict[str, Any]], List[List[Any]]]:
    """Load CSV/TSV/SSV format with optional SQL query support

    Returns empty list for empty input instead of raising an error.
    """
    # For SSV with columns parameter, use special handler
    if sep == r'\s+' and 'columns' in kwargs:
        columns = kwargs['columns']
        rest_kwargs = {k: v for k, v in kwargs.items() if k != 'columns'}
        return _load_ssv_with_columns(source, columns, has_header, **rest_kwargs)

    comment_mode = kwargs.pop('comment', 'line')
    if comment_mode is False:
        comment_mode = 'none'

    read_kwargs = _build_csv_read_kwargs(sep)
    if comment_mode == 'inline':
        read_kwargs['comment'] = '#'
        source = StringIO(_read_source(source, comment='none'))
    else:
        source = StringIO(_read_source(source, comment=comment_mode))

    try:
        if has_header:
            df = pd.read_csv(source, **read_kwargs)
            if 'query' in kwargs:
                return _apply_sql_query(df, kwargs['query'])
            return df.to_dict('records')
        else:
            df = pd.read_csv(source, header=None, **read_kwargs)
            # Strip trailing NaN values from each row (caused by ragged input)
            result = []
            for row in df.values.tolist():
                while row and (row[-1] is None or
                               (isinstance(row[-1], float) and
                                pd.isna(row[-1]))):
                    row.pop()
                result.append(row)
            return result
    except pd.errors.EmptyDataError:
        # pandas raises EmptyDataError for empty or whitespace-only input
        return []

# ─────────────────────────────────────────────────────────────────────────────
# Loader Dispatch Table

LOADERS = {
    'json': _load_json,
    'yaml': _load_yaml,
    'toml': _load_toml,
    'sqlite': _load_sqlite,
    'excel': _load_excel,
    'lines': _load_lines,
    'tsv': partial(_load_csv, sep='\t'),
    'ssv': partial(_load_csv, sep=r'\s+'),
    'csv': _load_csv,
}

# ─────────────────────────────────────────────────────────────────────────────
# Public API

def guess_format_from_filename(filename: str) -> str:
    """Guess data format from filename extension"""
    return FORMAT_EXTENSIONS.get(Path(filename).suffix.lower(), DEFAULT_FORMAT)

def _normalize_data_source(
    value: Union[str, Dict[str, Any]],
    table_name: str,
    data_format: Optional[str] = None,
) -> Tuple[Union[str, StringIO], str, Dict[str, Any]]:
    """Normalize data source specification to (source, format, load_kwargs) tuple.

    Path validation is not performed here; load_data() handles it via
    validate_file_path().
    """
    if isinstance(value, dict):
        if 'data' in value:
            # Inline data: {data: "...", format: "csv"}
            return StringIO(value['data']), value.get('format', DEFAULT_FORMAT), {}
        elif 'file' in value:
            # File with parameters: {file: "path.xlsx", table: "Sheet1", skiprows: "年"}
            filepath = value['file']
            fmt = value.get('format') or data_format or guess_format_from_filename(filepath)
            extra = {k: v for k, v in value.items() if k not in ('file', 'format')}
            return filepath, fmt, extra
        else:
            raise ValueError(
                f"Inline data dict for table '{table_name}' must have 'data' or 'file' key"
            )
    elif isinstance(value, str) and '\n' in value:
        return StringIO(value), DEFAULT_FORMAT, {}
    else:
        file_format = data_format or guess_format_from_filename(value)
        return value, file_format, {}

def load_data(
    source: Union[str, StringIO],
    format: Optional[str] = None,
    has_header: bool = True,
    **kwargs: Any
) -> Union[List[Any], Dict[str, Any]]:
    """Load data from file or StringIO

    Uses a dispatch table to delegate to format-specific loaders.

    Args:
        source: File path, StringIO object, or '-' for stdin
        format: Data format (csv, tsv, ssv, json, yaml, toml, sqlite, excel, lines)
               If None, auto-detect from filename
        has_header: Whether CSV/TSV/SSV has header row
        **kwargs: Format-specific options (e.g., table/query for sqlite)

    Returns:
        Loaded data (list or dict depending on format)
    """
    from .config import validate_file_path

    # Handle stdin
    if isinstance(source, str) and source == '-':
        source = StringIO(sys.stdin.read())
    elif isinstance(source, str):
        source = validate_file_path(source)

    # Normalize format aliases
    if format == 'spaces':
        format = 'ssv'

    # Auto-detect format
    if format is None:
        format = guess_format_from_filename(source) if isinstance(source, str) else DEFAULT_FORMAT

    # Dispatch to appropriate loader
    loader = LOADERS.get(format, LOADERS[DEFAULT_FORMAT])
    return loader(source, has_header=has_header, **kwargs)

# ─────────────────────────────────────────────────────────────────────────────
# Multi-table Support

def _is_resolved_data(value: Any) -> bool:
    """Check if value is already resolved variable data (not inline data spec)."""
    if isinstance(value, list):
        return True
    # Inline data dicts have 'data' key; file dicts have 'file' key
    return isinstance(value, dict) and 'data' not in value and 'format' not in value and 'file' not in value


def _to_dataframe(value: Any) -> pd.DataFrame:
    """Convert loaded data to DataFrame for SQL queries."""
    if isinstance(value, list):
        return pd.DataFrame(value)
    if isinstance(value, dict):
        return pd.DataFrame(list(value.values()))
    return pd.DataFrame()


def _query_tables(
    data_file: Dict[str, Any],
    data_format: Optional[str],
    has_header: bool,
    query: str
) -> List[Dict[str, Any]]:
    """Load multiple tables and execute SQL query"""
    datasets = _load_tables(data_file, data_format, has_header, {})
    tables = {name: _to_dataframe(data) for name, data in datasets.items()}
    return _apply_sql_query_multi(tables, query)

def _load_tables(
    data_file: Dict[str, Any],
    data_format: Optional[str],
    has_header: bool,
    load_kwargs: Dict[str, Any]
) -> Dict[str, Any]:
    """Load multiple tables for direct access"""
    datasets = {}
    for table_name, value in data_file.items():
        if _is_resolved_data(value):
            datasets[table_name] = value
            continue
        source, file_format, extra_kwargs = _normalize_data_source(
            value, table_name, data_format
        )
        merged_kwargs = {**load_kwargs, **extra_kwargs}
        datasets[table_name] = load_data(
            source,
            format=file_format,
            has_header=has_header,
            **merged_kwargs
        )
    return datasets

def _load_embedz_data(
    data_file: Optional[Union[str, Dict[str, Any]]],
    data_part: Optional[str],
    config: Dict[str, Any],
    data_format: Optional[str],
    has_header: bool,
    load_kwargs: Dict[str, Any]
) -> Union[List[Any], Dict[str, Any]]:
    """Load data from file(s), inline data, or multi-table sources"""
    if data_file and data_part:
        raise ValueError(
            "Cannot specify both 'data' attribute and inline data. "
            "Use either 'data: filename.csv' or provide inline data after '---', not both."
        )

    if data_file:
        if isinstance(data_file, dict):
            if config.get('query'):
                return _query_tables(data_file, data_format, has_header, config['query'])
            else:
                return _load_tables(data_file, data_format, has_header, load_kwargs)
        else:
            return load_data(data_file, format=data_format, has_header=has_header, **load_kwargs)

    if data_part:
        return load_data(StringIO(data_part), format=data_format or DEFAULT_FORMAT, has_header=has_header, **load_kwargs)

    return []
